/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/RestoreSmBackupCommand.java#4 $
 *
 * Copyright (c) 2016 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.snapcenter.cli.command;

import static com.netapp.snapcenter.cli.command.validation.SmsBackupValidationHelper.ORACLE_DATA_BACKUP_TYPE;

import java.util.List;
import java.util.Map;

import org.apache.commons.lang3.StringUtils;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.Parameters;
import com.beust.jcommander.ParametersDelegate;
import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.smcore.plugin.contracts.SmGetSMTPServerResponse;
import com.netapp.smcore.plugin.contracts.SmRecoveryRequest;
import com.netapp.smcore.plugin.contracts.SmRecoveryResponse;
import com.netapp.smcore.plugin.contracts.SmRequest;
import com.netapp.smcore.plugin.contracts.SmResponse;
import com.netapp.snapcenter.cli.command.parameters.NotificationParameters;
import com.netapp.snapcenter.cli.command.parameters.PrePostScriptParameters;
import com.netapp.snapcenter.cli.command.parameters.SDURestoreParameters;
import com.netapp.snapcenter.cli.command.parameters.SMORestoreParameters;
import com.netapp.snapcenter.cli.command.parser.CustomStringToMapConverter;
import com.netapp.snapcenter.cli.command.parser.NoParameterSplitter;
import com.netapp.snapcenter.cli.command.validation.GetRecentBackupInformationRequest;
import com.netapp.snapcenter.cli.command.validation.GetRecentBackupInformationResponse;
import com.netapp.snapcenter.cli.command.validation.RestoreCommandParametersValidationUtil;
import com.netapp.snapcenter.cli.command.validation.SmsBackupValidationHelper;
import com.netapp.snapcenter.cli.command.validation.SmsCommandArgumentsValidationUtil;
import com.netapp.snapcenter.cli.command.validation.SmsCommandRequestValidationHelper;
import com.netapp.snapcenter.cli.contracts.PluginCode;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;
import com.netapp.snapcenter.cli.utils.ProgressBar;
import com.netapp.snapcenter.cli.utils.ProgressBarUtil;
import com.netapp.spl.common.restclient.config.SMSRestConfiguration;
import com.netapp.spl.common.restclient.context.SMSRestInvocationContext;

/**
 * This class is used to perform the Restore-SmBackup command functionality. It
 * will contain plugin specific command parameters
 *
 */
@Parameters(resourceBundle = "snapcenter-cli", commandDescriptionKey = "cli.restoresmbackup.command.description")
public class RestoreSmBackupCommand extends AbstractSMSCommand {

    private ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(RestoreSmBackupCommand.class);

    private static final String COMMANDNAME = MessageBundle
            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_COMMAND_NAME);

    @Parameter(names = { "-PluginCode" }, descriptionKey = "cli.restoresmbackup.command.pluginCode.option.description")
    private PluginCode pluginCode = PluginCode.SCO;

    @Parameter(names = { "-AppObjectId" }, descriptionKey = "cli.restoresmbackup.command.appObjectId.option.description", required = true)
    private String appObjectId;

    @Parameter(names = { "-BackupId" }, descriptionKey = "cli.restoresmbackup.command.backupId.option.description")
    private String backupId;

    @Parameter(names = { "-BackupName" }, descriptionKey = "cli.restoresmbackup.command.backupName.option.description")
    private String backupName;

    @Parameter(names = { "-RestoreLastBackup" }, descriptionKey = "cli.restoresmbackup.command.restoreLastBackup.option.description")
    private Integer restoreLastBackup;

    @Parameter(names = { "-WaitForCompletion" }, descriptionKey = "cli.restoresmbackup.command.waitForCompletion.option.description")
    private boolean waitForCompletion;

    @ParametersDelegate
    SMORestoreParameters smoRestoreParameters = new SMORestoreParameters();

    @ParametersDelegate
    SDURestoreParameters sduRestoreParameters = new SDURestoreParameters();

    @Parameter(names = { "-SecondaryLocator" }, descriptionKey = "cli.restoresmbackup.command.secondaryLocator.option.description", converter = CustomStringToMapConverter.class, splitter = NoParameterSplitter.class)
    private List<Map<String, String>> secondaryLocators;

    @ParametersDelegate
    private PrePostScriptParameters prePostScriptParameters = new PrePostScriptParameters();

    @ParametersDelegate
    private NotificationParameters notificationParameters = new NotificationParameters();

    @Override
    public String getCommandName() {
        return COMMANDNAME;
    }

    @Override
    protected void processCommandArguments() throws CLIException {
        boolean restoreFromSecondary = false;

        SmsBackupValidationHelper smsBackupValidationHelper = new SmsBackupValidationHelper(
                getSnapCenterServerServiceHelper());
        if (StringUtils.isBlank(backupId) && StringUtils.isBlank(backupName)
                && restoreLastBackup == null) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_BACKUP_INFORMATION_NOT_SPECIFIED));
        }

        if (StringUtils.isNotBlank(backupId)
                && StringUtils.isNotBlank(backupName)
                || StringUtils.isNotBlank(backupName)
                && restoreLastBackup != null || restoreLastBackup != null
                && StringUtils.isNotBlank(backupId)) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_MORE_THAN_ONE_BACKUP_INFORMATION_SPECIFIED));
        }

        if (restoreLastBackup != null && restoreLastBackup < 0) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_RESTORE_LAST_BACKUP_VALUE_CANNOT_BE_NEGATIVE));
        }

        if (!SmsCommandArgumentsValidationUtil.isValidAppObjectId(appObjectId)) {
            throw new CLIException(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_COMMAND_PARAMETER_APP_OBJECT_ID_NOT_VALID));
        }

        if (secondaryLocators != null && secondaryLocators.size() > 0) {
            SmsCommandArgumentsValidationUtil
                    .validateSecondaryLocators(secondaryLocators);
            restoreFromSecondary = true;
            CLIUtils.writeToLogAndConsole(
                    MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_WILL_BE_PERFORMED_FROM_SECONDARY),
                    logger, LogLevel.INFO);
        }

        switch (getPluginCode()) {
            case SCO:
                RestoreCommandParametersValidationUtil
                        .processAndValidateSMORestoreParameters(smoRestoreParameters, super.getSnapCenterServerServiceHelper());
                // TODO: Revisit this when SCO in Windows/ SCU going standalone
                RestoreCommandParametersValidationUtil
                        .processAndValidateSDURestoreParameters(sduRestoreParameters);

                if (restoreLastBackup != null) {
                    GetRecentBackupInformationRequest request = new GetRecentBackupInformationRequest();
                    request.setAppObjectId(getAppObjectId());
                    request.setSecondary(restoreFromSecondary);
                    request.setRecentBackupNumber(restoreLastBackup);
                    request.setPluginCode(getPluginCode());
                    request.setBackupType(ORACLE_DATA_BACKUP_TYPE);

                    String lastBackupName = getLastOracleDataBackupName(
                            smsBackupValidationHelper, request);
                    CLIUtils.writeToLogAndConsole(
                            MessageBundle
                                    .localizeMessage(
                                            CLIMsgKey.CLI_RESTORESMBACKUP_RESTORE_LAST_BACKUP_OPTION_IS_CHOSEN,
                                            restoreLastBackup,
                                            lastBackupName), logger,
                            LogLevel.INFO);
                    setBackupName(lastBackupName);
                } else if (!isValidOracleDataBackup(smsBackupValidationHelper,
                        restoreFromSecondary)) {
                    String errorMessage = null;
                    if (restoreFromSecondary) {
                        errorMessage = StringUtils.isBlank(getBackupName()) ? MessageBundle
                                .localizeMessage(
                                        CLIMsgKey.CLI_RESTORESMBACKUP_NOT_VALID_ORACLE_SECONDARY_BACKUP_ID_OR_RESOURCE_NOT_IN_BACKUP,
                                        getBackupId(), getAppObjectId())
                                : MessageBundle
                                        .localizeMessage(
                                                CLIMsgKey.CLI_RESTORESMBACKUP_NOT_VALID_ORACLE_SECONDARY_BACKUP_NAME_OR_RESOURCE_NOT_IN_BACKUP,
                                                getBackupName(),
                                                getAppObjectId());
                        throw new CLIException(errorMessage);
                    } else {
                        errorMessage = StringUtils.isBlank(getBackupName()) ? MessageBundle
                                .localizeMessage(
                                        CLIMsgKey.CLI_RESTORESMBACKUP_NOT_VALID_ORACLE_BACKUP_ID_OR_RESOURCE_NOT_IN_BACKUP,
                                        getBackupId(), getAppObjectId())
                                : MessageBundle
                                        .localizeMessage(
                                                CLIMsgKey.CLI_RESTORESMBACKUP_NOT_VALID_ORACLE_BACKUP_NAME_OR_RESOURCE_NOT_IN_BACKUP,
                                                getBackupName(),
                                                getAppObjectId());

                        throw new CLIException(errorMessage);
                    }
                }
                break;
            default:
                throw new CLIException(MessageBundle.localizeMessage(
                        CLIMsgKey.CLI_COMMAND_NOT_SUPPORTED_FOR_PLUGIN,
                        pluginCode.name()));
        }

        SmsCommandArgumentsValidationUtil
                .processAndValidatePrePostScriptParameters(
                        prePostScriptParameters, pluginCode);

		SmsCommandRequestValidationHelper smsCommandRequestValidationHelper = new SmsCommandRequestValidationHelper(
				super.getSnapCenterServerServiceHelper());
		SmGetSMTPServerResponse smGetSMTPServerResponse = smsCommandRequestValidationHelper
				.getSmSMTPServerDetailsFromServer();

        SmsCommandArgumentsValidationUtil
                .validateAndSetNotificationParameters(notificationParameters,smGetSMTPServerResponse,COMMANDNAME);

        CLIUtils.promptUserToContinue(MessageBundle
                .localizeMessage(CLIMsgKey.CLI_RESTORESMBACKUP_PROMPT_USER_RESTORE_BACKUP));
    }

    private String getLastOracleDataBackupName(
            SmsBackupValidationHelper smsBackupValidationHelper,
            GetRecentBackupInformationRequest req) throws CLIException {
        String lastBackupName = null;
        ProgressBar progressBar = ProgressBarUtil
                .createAndStartProgressBar("Fetching last backup ");
        try {
            GetRecentBackupInformationResponse resp = smsBackupValidationHelper
                    .getRecentOracleDataBackupInformation(req);
            if (resp.getNumberOfBackups() == 0) {
                String errorMessage = null;
                if (req.isSecondary()) {
                    errorMessage = MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_RESTORESMBACKUP_RESTORE_LAST_BACKUP_NO_ORACLE_SECONDARY_DATA_BACKUP_EXISTS_FOR_RESOURCE,
                                    req.getAppObjectId());
                } else {
                    errorMessage = MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_RESTORESMBACKUP_RESTORE_LAST_BACKUP_NO_ORACLE_DATA_BACKUP_EXISTS_FOR_RESOURCE,
                                    req.getAppObjectId());
                }
                throw new CLIException(errorMessage);
            } else if (resp.getRecentBackup() == null) {
                String errorMessage = null;
                if (req.isSecondary()) {
                    errorMessage = MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_RESTORESMBACKUP_RESTORE_LAST_BACKUP_NUMBER_OF_LAST_BACKUP_IS_MORE_THAN_EXISTSING_SECONDARY_BACKUPS,
                                    req.getRecentBackupNumber(),
                                    req.getAppObjectId(),
                                    resp.getNumberOfBackups());
                } else {
                    errorMessage = MessageBundle
                            .localizeMessage(
                                    CLIMsgKey.CLI_RESTORESMBACKUP_RESTORE_LAST_BACKUP_NUMBER_OF_LAST_BACKUP_IS_MORE_THAN_EXISTSING_BACKUPS,
                                    req.getRecentBackupNumber(),
                                    req.getAppObjectId(),
                                    resp.getNumberOfBackups());
                }
                throw new CLIException(errorMessage);
            } else {
                lastBackupName = resp.getRecentBackup().getBackupName();
            }
        } catch (CLIException e) {
            throw e;
        } finally {
            ProgressBarUtil.stopProgressBar(progressBar);
        }
        return lastBackupName;
    }

    private boolean isValidOracleDataBackup(
            SmsBackupValidationHelper smsBackupValidationHelper,
            boolean restoreFromSecondary) throws CLIException {
        boolean validOracleDataBackup = false;
        ProgressBar progressBar = ProgressBarUtil
                .createAndStartProgressBar("Validating backup details ");
        try {
            // Validate the specified backup
            validOracleDataBackup = smsBackupValidationHelper
                    .checkIfValidOracleDataBackup(backupId, backupName,
                            appObjectId, restoreFromSecondary);
        } catch (CLIException e) {
            throw e;
        } finally {
            ProgressBarUtil.stopProgressBar(progressBar);
        }
        return validOracleDataBackup;
    }

    @Override
    protected <T extends SmRequest> SMSRestInvocationContext<?, ?> getCommandSpecificSMSRestInvocationContext(
            T cmdRequest) {
        SMSRestInvocationContext<SmRecoveryRequest, SmRecoveryResponse> smsRestInvocationContext = new SMSRestInvocationContext<>();
        smsRestInvocationContext.setRequest((SmRecoveryRequest) cmdRequest);
        smsRestInvocationContext.setResponseClass(SmRecoveryResponse.class);
        smsRestInvocationContext
                .setServiceUrlPath(SMSRestConfiguration.SMS_RESTORE_BACKUP_SERVICE_PATH);
        return smsRestInvocationContext;
    }

    @Override
    protected <T> void pollForJobStatus(SmResponse smResponse)
            throws CLIException {
        doPollForJobStatus(smResponse, isWaitForCompletion());
    }

    public PluginCode getPluginCode() {
        return pluginCode;
    }

    public void setPluginCode(PluginCode pluginCode) {
        this.pluginCode = pluginCode;
    }

    public String getAppObjectId() {
        return appObjectId;
    }

    public void setAppObjectId(String appObjectId) {
        this.appObjectId = appObjectId;
    }

    public String getBackupId() {
        return backupId;
    }

    public void setBackupId(String backupId) {
        this.backupId = backupId;
    }

    public String getBackupName() {
        return backupName;
    }

    public void setBackupName(String backupName) {
        this.backupName = backupName;
    }

    public SMORestoreParameters getSmoRestoreParameters() {
        return smoRestoreParameters;
    }

    public void setSmoRestoreParameters(
            SMORestoreParameters smoRestoreParameters) {
        this.smoRestoreParameters = smoRestoreParameters;
    }

    public PrePostScriptParameters getPrePostScriptParameters() {
        return prePostScriptParameters;
    }

    public void setPrePostScriptParameters(
            PrePostScriptParameters prePostScriptParameters) {
        this.prePostScriptParameters = prePostScriptParameters;
    }

    public void setSduRestoreParameters(
            SDURestoreParameters sduRestoreParameters) {
        this.sduRestoreParameters = sduRestoreParameters;
    }

    public NotificationParameters getNotificationParameters() {
        return notificationParameters;
    }

    public void setNotificationParameters(
            NotificationParameters notificationParameters) {
        this.notificationParameters = notificationParameters;
    }

    public SDURestoreParameters getSduRestoreParameters() {
        return sduRestoreParameters;
    }

    public void setWaitForCompletion(boolean waitForCompletion) {
        this.waitForCompletion = waitForCompletion;
    }

    public boolean isWaitForCompletion() {
        return waitForCompletion;
    }

    public void setSecondaryLocators(List<Map<String, String>> secondaryLocators) {
        this.secondaryLocators = secondaryLocators;
    }

    public List<Map<String, String>> getSecondaryLocators() {
        return secondaryLocators;
    }

    public Integer getRestoreLastBackup() {
        return restoreLastBackup;
    }

    public void setRestoreLastBackup(Integer restoreLastBackup) {
        this.restoreLastBackup = restoreLastBackup;
    }
}