/*
 * $Id: //depot/prod/spl/main/snapcenter-cli/src/main/java/com/netapp/snapcenter/cli/command/validation/OracleCloneSpecificationValidationUtil.java#1 $
 * 
 * Copyright (c) 2015 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.snapcenter.cli.command.validation;

import static com.netapp.snapcenter.cli.utils.CLIUtils.SEMICOLON;

import java.util.HashMap;
import java.util.List;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;

import com.netapp.common.logging.MessageBundle;
import com.netapp.smcore.loader.logger.ExtendedGenericLogger;
import com.netapp.smcore.loader.logger.LoggerFactory;
import com.netapp.snapcenter.cli.clone.contracts.AsmAuthentication;
import com.netapp.snapcenter.cli.clone.contracts.ControlFile;
import com.netapp.snapcenter.cli.clone.contracts.ControlFilesConfiguration;
import com.netapp.snapcenter.cli.clone.contracts.DataFilesConfiguration;
import com.netapp.snapcenter.cli.clone.contracts.DataFilesLocation;
import com.netapp.snapcenter.cli.clone.contracts.DatabaseAuthentication;
import com.netapp.snapcenter.cli.clone.contracts.DatabaseParameter;
import com.netapp.snapcenter.cli.clone.contracts.DatabaseParameters;
import com.netapp.snapcenter.cli.clone.contracts.DatabaseSpecification;
import com.netapp.snapcenter.cli.clone.contracts.OracleCloneSpecification;
import com.netapp.snapcenter.cli.clone.contracts.OracleOsAccount;
import com.netapp.snapcenter.cli.clone.contracts.PdbConfigurationType;
import com.netapp.snapcenter.cli.clone.contracts.PdbsConfiguration;
import com.netapp.snapcenter.cli.clone.contracts.Redogroup;
import com.netapp.snapcenter.cli.clone.contracts.RedologFile;
import com.netapp.snapcenter.cli.clone.contracts.RedologsConfiguration;
import com.netapp.snapcenter.cli.clone.contracts.SecondaryConfiguration;
import com.netapp.snapcenter.cli.clone.contracts.SecondaryLocator;
import com.netapp.snapcenter.cli.clone.contracts.SecondaryLocators;
import com.netapp.snapcenter.cli.clone.contracts.SqlStatements;
import com.netapp.snapcenter.cli.clone.contracts.StorageSpecification;
import com.netapp.snapcenter.cli.exception.CLIException;
import com.netapp.snapcenter.cli.logging.CLIMsgKey;
import com.netapp.snapcenter.cli.utils.CLIUtils;
import com.netapp.snapcenter.cli.utils.CLIUtils.LogLevel;

/**
 * This class Oracle clone specification file related validation.
 *
 */
public class OracleCloneSpecificationValidationUtil {

    public static final Pattern SID_PATTERN = Pattern.compile("[A-Za-z][A-Za-z0-9]{0,}");

    private OracleCloneSpecificationValidationUtil() {

    }

    private static ExtendedGenericLogger logger = LoggerFactory
            .getExtendedClassAwareLogger(OracleCloneSpecificationValidationUtil.class);

    public static void validateCloneSpec(
            OracleCloneSpecification oracleCloneSpecification, String commandName)
            throws CLIException {
        if (oracleCloneSpecification == null
                || oracleCloneSpecification.getDatabaseSpecification() == null
                || oracleCloneSpecification.getStorageSpecification() == null) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_INVALID_SPEC));
        }
        // Skip Backupname Check in case of Refresh-SmClone command since it does not use Backupname from CloneSpecificationFile
        if (StringUtils.isBlank(oracleCloneSpecification.getBackupname()) && !commandName.equals(MessageBundle.localizeMessage(CLIMsgKey.CLI_REFRESHSMCLONE_COMMAND_NAME))) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_BACKUPNAME_EMPTY));
        }
        if (StringUtils.isBlank(oracleCloneSpecification.getAppobjectid())) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_APPOBJECTID_EMPTY));
        }
        if (StringUtils.isBlank(oracleCloneSpecification.getCloneDatabaseSid())) {
        	throw new CLIException(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_CLONE_DATABASE_SID_EMPTY));
        }
        if (!SID_PATTERN.matcher(oracleCloneSpecification.getCloneDatabaseSid()).matches()) {
        	throw new CLIException(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_CLONE_DATABASE_SID_INVALID,
                            oracleCloneSpecification.getCloneDatabaseSid()));
        }
        if (oracleCloneSpecification.getCloneDatabaseSid().length() > 8) {
            throw new CLIException(MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_CREATESMORACLECLONESPECIFICATION_CLONE_DATABASE_SID_IS_MORE_THAN_EIGHT_CHARACTERS));
        }

        String[] appObjectSplit = StringUtils.split(oracleCloneSpecification.getAppobjectid(), "\\");
        if (appObjectSplit[1].equals(oracleCloneSpecification.getCloneDatabaseSid())) {
            throw new CLIException(MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_CREATESMORACLECLONESPECIFICATION_CLONE_DATABASE_SID_CANNOT_BE_SAME_AS_SOURCE_DB));
        }
        validateStorageSpecification(oracleCloneSpecification
                .getStorageSpecification());
        validateDatabaseSpecification(oracleCloneSpecification
                .getDatabaseSpecification());
    }

    private static void validateStorageSpecification(
            StorageSpecification storageSpecification) throws CLIException {
        if (storageSpecification.getStorageMapping() == null) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_STORAGE_MAPPING_NOT_SPECIFIED));
        }
        DataFilesConfiguration dataFilesConfiguration = storageSpecification
                .getStorageMapping().getDataFilesConfiguration();
        if (dataFilesConfiguration == null
                || dataFilesConfiguration.getDataFilesLocations() == null
                || dataFilesConfiguration.getDataFilesLocations()
                        .getDataFilesLocation().isEmpty()) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_MOUNT_POINTS_CONFIGURATION_NOT_SPECIFIED));
        }
        for (DataFilesLocation dataFilesLocation : dataFilesConfiguration
                .getDataFilesLocations().getDataFilesLocation()) {
            if (StringUtils.isBlank(dataFilesLocation.getSourceLocation())
                    || StringUtils
                            .isBlank(dataFilesLocation.getCloneLocation())
                    || dataFilesLocation.getLocationType() == null) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_INVALID_DATAFILE_LOCATION));
            }
        }

        HashMap<String, String> primaryToSecondaryMap = new HashMap<>();

        /**
         * Validate secondary configuration for logs.
         */
        SecondaryConfiguration secondaryConfigurationForArchivelogs = storageSpecification
                .getStorageMapping().getSecondaryConfigurationArchivelogs();
        if (secondaryConfigurationForArchivelogs != null) {
            SecondaryLocators secondaryLocators = secondaryConfigurationForArchivelogs
                    .getSecondaryLocators();
            if (secondaryLocators != null) {
                for (SecondaryLocator sl : secondaryLocators
                        .getSecondaryLocator()) {
                    validateSecondaryLocator(primaryToSecondaryMap, sl);
                }
            }
        }

        primaryToSecondaryMap.clear();

        /**
         * Validate secondary configuration for data.
         */
        SecondaryConfiguration secondaryConfigurationForDatafiles = storageSpecification
                .getStorageMapping().getSecondaryConfigurationDatafiles();
        if (secondaryConfigurationForDatafiles != null) {
            SecondaryLocators secondaryLocators = secondaryConfigurationForDatafiles
                    .getSecondaryLocators();
            if (secondaryLocators != null) {
                for (SecondaryLocator sl : secondaryLocators
                        .getSecondaryLocator()) {
                    validateSecondaryLocator(primaryToSecondaryMap, sl);
                }
                CLIUtils.writeToLogAndConsole(
                        MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_CLONE_OPERATION_WILL_BE_PERFORMED_FROM_SECONDARY),
                        logger, LogLevel.INFO);
            }
        }
    }

    private static void validateSecondaryLocator(
            HashMap<String, String> primaryToSecondaryMap, SecondaryLocator sl) throws CLIException {
        if (sl.getPrimary() == null || sl.getSecondary() == null) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_INVALID_SECONDARY_LOCATORS_CONFIGURATION_ERROR));
        }
        if (StringUtils.isBlank(sl.getPrimary().getSvm())
                || StringUtils.isBlank(sl.getPrimary().getVolume())) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_PRIMARY_SVM_VOLUME_NOT_SPECIFIED_ERROR));
        }
        if (StringUtils.isBlank(sl.getSecondary().getSvm())
                || StringUtils.isBlank(sl.getSecondary()
                        .getVolume())) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_SECONDARY_SVM_VOLUME_NOT_SPECIFIED_ERROR));
        }
        if (primaryToSecondaryMap.containsKey(String.format(
                "%s:%s", sl.getPrimary().getSvm(), sl.getPrimary()
                        .getVolume()))) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(
                            CLIMsgKey.CLI_ORACLE_CLONESPEC_MULIPLE_PRIMARY_TO_SECONDARY_LOCATOR_ERROR,
                            String.format("%s:%s", sl.getPrimary()
                                    .getSvm(), sl.getPrimary()
                                    .getVolume())));
        } else {
            primaryToSecondaryMap.put(String.format("%s:%s", sl
                    .getPrimary().getSvm(), sl.getPrimary()
                    .getVolume()), String.format("%s:%s", sl
                    .getSecondary().getSvm(), sl.getSecondary()
                    .getVolume()));
        }        
    }

    private static void validateDatabaseSpecification(
            DatabaseSpecification databaseSpecification) throws CLIException {

        if (databaseSpecification.getDatabaseAuthentication() != null) {
            DatabaseAuthentication dbAuth = databaseSpecification
                    .getDatabaseAuthentication();
            if (StringUtils.isBlank(dbAuth.getRunasName())
                    && !isValidPort(dbAuth.getPort())) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_INVALID_DB_AUTH_PORT));
            }
        }

        if (databaseSpecification.getAsmAuthentication() != null) {
            AsmAuthentication asmAuth = databaseSpecification
                    .getAsmAuthentication();
            if (StringUtils.isNotBlank(asmAuth.getRunasName())
                    && !isValidPort(asmAuth.getPort())) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_INVALID_ASM_AUTH_PORT));
            }
        }

        ControlFilesConfiguration controlFilesConfiguration = databaseSpecification
                .getControlFilesConfiguration();
        if (controlFilesConfiguration == null
                || controlFilesConfiguration.getControlFiles() == null
                || controlFilesConfiguration.getControlFiles().getControlFile()
                        .isEmpty()) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_CONTROL_FILES_CONFIGURATION_NOT_SPECIFIED));
        } else {
            List<ControlFile> controlFiles = controlFilesConfiguration
                    .getControlFiles().getControlFile();
            for (ControlFile cf : controlFiles) {
                if (StringUtils.isBlank(cf.getPath())) {
                    constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_CONTROL_FILE_PATH_EMPTY_ERROR));
                }
            }
        }

        RedologsConfiguration redologsConfiguration = databaseSpecification
                .getRedologsConfiguration();
        if (redologsConfiguration == null
                || redologsConfiguration.getRedologs() == null
                || redologsConfiguration.getRedologs().getRedogroup().isEmpty()) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_REDO_LOGS_CONFIGURATION_NOT_SPECIFIED));
        } else {
            List<Redogroup> redogroups = redologsConfiguration.getRedologs()
                    .getRedogroup();
            validateRedoGroups(redogroups);
        }

        if (databaseSpecification.getSqlStatements() != null) {
            validateSqlStatements(databaseSpecification.getSqlStatements());
        }

        if (databaseSpecification.getOracleOsAccount() != null) {
            OracleOsAccount oracleOsAccount = databaseSpecification
                    .getOracleOsAccount();
            if (StringUtils.isBlank(oracleOsAccount.getUsername())) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_OS_USERNAME_EMPTY_ERROR));
            }
            if (StringUtils.isBlank(oracleOsAccount.getGroup())) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_OS_GROUPNAME_EMPTY_ERROR));
            }
        } else {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_OS_ACCOUNT_DETAILS_NOT_SPECIFIED));
        }

        if (StringUtils.isBlank(databaseSpecification.getOracleVersion())) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_VERSION_EMPTY_ERROR));
        }

        if (StringUtils.isBlank(databaseSpecification.getOracleHome())) {
            constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                    .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_HOME_EMPTY_ERROR));
        }

        if (databaseSpecification.getDatabaseParameters() != null) {
            validateDatabaseParameters(databaseSpecification
                    .getDatabaseParameters());
        }

        if (databaseSpecification.getPdbsConfiguration() != null) {
            validatePdbsConfiguration(databaseSpecification
                    .getPdbsConfiguration());
        }

    }

    private static boolean isValidPort(Integer port) {
        if (port == null || port < 0 || port > 65536) {
            return false;
        }
        return true;
    }

    private static void validatePdbsConfiguration(
            PdbsConfiguration pdbsConfiguration) throws CLIException {
        for (PdbConfigurationType pdb : pdbsConfiguration.getPdbConfiguration()) {
            if (StringUtils.isBlank(pdb.getCurrentName())) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_PDB_CONFIGURATION_CURRENT_NAME_CANNOT_BE_EMPTY));
            }
            if (StringUtils.isNotBlank(pdb.getNewName())
                    && !SID_PATTERN.matcher(pdb.getNewName()).matches()) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(
                                CLIMsgKey.CLI_ORACLE_CLONESPEC_CLONE_PDB_NAME_INVALID,
                                pdb.getNewName()));
            }
        }
    }

    private static void validateDatabaseParameters(
            DatabaseParameters databaseParameters) throws CLIException {
        for (DatabaseParameter dbParam : databaseParameters
                .getDatabaseParameter()) {
            if (StringUtils.isBlank(dbParam.getName())) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_DATABASE_PARAMETER_NAME_CANNOT_BE_EMPTY));
            }
        }
    }

    private static void validateSqlStatements(SqlStatements sqlStatements)
            throws CLIException {
        for (String stmt : sqlStatements.getSqlStatement()) {
            if (StringUtils.isNotBlank(stmt)
                    && !stmt.trim().endsWith(SEMICOLON)) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_ORACLE_SQL_STATEMENT_INVALID));
            }
        }
    }

    private static void validateRedoGroups(List<Redogroup> redogroups)
            throws CLIException {
        if (!redogroups.isEmpty()) {
            if (redogroups.size() < 3) {
                constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                        .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_REDOGROUPS_SHOULD_BE_MINIMUM_THREE));
            }
            for (Redogroup rg : redogroups) {

                if (rg.getTotalSize().getValue() <= 0) {
                    constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_REDOGROUP_TOTAL_SIZE_SHOULD_BE_GREATER_THAN_ZERO));
                }
                if (rg.getNumber() <= 0) {
                    constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_REDOGROUP_NUMBER_SHOULD_BE_GREATER_THAN_ZERO));
                }
                if (rg.getRedologFiles() == null
                        || rg.getRedologFiles().getRedologFile().isEmpty()) {
                    constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                            .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_REDOGROUP_SHOULD_HAVE_ATLEAST_ONE_REDOLOG_FILE));

                }
                for (RedologFile redoFile : rg.getRedologFiles()
                        .getRedologFile()) {
                    if (redoFile == null
                            || StringUtils.isEmpty(redoFile.getPath())) {
                        constructAndThrowExceptionForOracleCloneSpecValidationError(MessageBundle
                                .localizeMessage(CLIMsgKey.CLI_ORACLE_CLONESPEC_REDOLOG_FILE_PATH_EMPTY_ERROR));
                    }
                }
            }
        }
    }

    private static void constructAndThrowExceptionForOracleCloneSpecValidationError(
            String errorMessage) throws CLIException {
        throw new CLIException(MessageBundle.localizeMessage(
                CLIMsgKey.CLI_ORACLE_CLONESPEC_VALIDATION_ERROR, errorMessage));

    }

}
